#!/bin/bash
########################################################################
#
# @(#)osvd_install.sh
#
# Copyright FUJITSU LIMITED 2016
#
# @version 1.4.0 (2016.01)
#
########################################################################

#---------------
# Definition
#---------------
PRODUCT_NAME="ETERNUS OpenStack VolumeDriver"
VERSION="1.4.0"
COPYRIGHT="Copyright FUJITSU LIMITED 2016"

SH_NAME=$(basename $0)
CAT_CMD="/bin/cat"
CP_CMD="/bin/cp"
GREP_CMD="/bin/grep"
LS_CMD="/bin/ls"
RM_CMD="/bin/rm"
SED_CMD="/bin/sed"
AWK_CMD="/usr/bin/awk"
CUT_CMD="/usr/bin/cut"
FIND_CMD="/usr/bin/find"
TR_CMD="/usr/bin/tr"
WHOAMI_CMD="/usr/bin/whoami"
MKDIR_CMD="/usr/bin/mkdir"
SSHKEYGEN_CMD="/usr/bin/ssh-keygen"

. ./osvd_install.def
CINDER_CONF="/etc/cinder/cinder.conf"

ENABLED_BACKENDS=""
VOLUME_BACKEND_NAME=""

#---------------
# Functions
#---------------

# print with level
PR()
{
	msg=$1
	level=$2
	if [ "${level}" = "" ]
	then
		level="INFO"
		echo "${SH_NAME}: ${level}: ${msg}"
	else
		echo "${SH_NAME}: ${level}: ${msg}" >&2
	fi
}

# print product information
PR_PRODUCT_INFO()
{
	echo ""
	echo ${PRODUCT_NAME}
	echo ${VERSION}
	echo ${COPYRIGHT}
	echo ""
}

# print error and usage
PR_ERROR_USAGE()
{
	echo "Usage: ./osvd_install.sh {install | update | uninstall} [ install_config_file ]"
	exit 1
}

# print title
TITLE()
{
	echo ""
	echo "====================================="
	echo "  $1"
	echo "====================================="
}

# find install directory
FIND_INSTDIR()
{
	CINDER_DIR=`python -c 'import cinder; print cinder.__file__'`
	inst_dir=`echo $CINDER_DIR | sed -e 's/__init__.pyc/volume\/drivers\//g'`
	if [ ! -d "${inst_dir}" ]; then
        PR "Can not find install directory." "ERROR"
		exit 1
	fi
	echo ${inst_dir}
}

#---------------
# INSTALL SubFunctions
#---------------

# check installed
CHECK_INSTALLED()
{
	local inst_dir=$1
	local inst_cfg=$2

	TITLE "check volume driver"

	if [ -d ${inst_dir}fujitsu ]
	then
		PR "This driver has already been installed."
		if [ -z "${inst_cfg}" ]; then
			YES_NO "Overwrite the driver?"
			if [ $? -ne 0 ]; then
				PR "The installation is ended."
				PR "Please execute \"${SH_NAME} uninstall\" to reinstall this driver."
				exit 0
			else
				DELETE_FILE ${inst_dir}
			fi
		else
			DELETE_FILE ${inst_dir}
		fi
	fi
}

# put files
PUT_FILE()
{
	inst_dir=$1
	TITLE "Put driver files"
	echo "Copy driver files to ${inst_dir}"
	${CP_CMD} -r ../../fujitsu ${inst_dir}
	if [ $? -ne 0 ]; then
		PR "install FAILED!!" "ERROR"
		exit 1
	fi
}

#---------------
# UPDATE_CONFIG SubFunctions
#---------------
# select protocol
SELECT_PROTOCOL()
{
	TITLE "select protocol"
	echo "1) FibreChannel"
	echo "2) iSCSI"
	echo -n "Which way of FibreChannel and iSCSI do you use?[1 or 2]: "
	read -r protocol
	while [ "${protocol}" != "2" ]
	do
		if [ "${protocol}" == "1" ]; then
			echo -n "This version supports only iSCSI. Please select 2:"
		else
			echo -n "Please select 1 or 2: "
		fi
		read -r protocol
	done

	case "${protocol}" in
		"1") PR "protocol is FibreChannel";;
		"2") PR "protocol is iSCSI";;
	esac
	return ${protocol}
}

# yes or not
YES_NO()
{
	while true
	do
		echo -n "$1 [yes/no]: "
		read -r answer
		case ${answer} in
			yes)
				return 0;;
			no)
				return 1;;
			*)
				echo "please input yes or no.";;
		esac
	done
}

# output information to cinder.conf
OUTPUT_CINDER_CONF()
{
	local section=$1
	local conf=$2
	local params=${*:3}
	local driver_path=""
	local i=0
	local line=""
	local line1=""
	local line2=""
	local line3=""

	${GREP_CMD} -i "\[${section}\]" ${CINDER_CONF} >/dev/null
	if [ $? -eq 0 ]; then
		echo "SECTION: ${section} has already existed, Skip update"
		return 1
	fi

	if [ ${prtcl} == 'fc' ]; then
		echo "SECTION: ${section}, Skip (Only iSCSI is supported)"
		return 1
	elif [ ${prtcl} == 'iscsi' ]; then
		driver_path="volume_driver=cinder.volume.drivers.fujitsu.extension.eternus_dx_iscsi.FJDXISCSIDriver"
	fi

	drvconfig_path="cinder_eternus_config_file=${conf}"

	# output to cinder.conf
	echo "[${section}]" >> ${CINDER_CONF}
	echo "${driver_path}" >> ${CINDER_CONF}
	echo "${drvconfig_path}" >> ${CINDER_CONF}

	line1=`${SED_CMD} -n '/^\s*enabled_backends\s*=/=' ${CINDER_CONF} | sed -n "1"p`
	line2=`${SED_CMD} -n '/^#enabled_backends\s*=/=' ${CINDER_CONF} | sed -n "1"p`
	line3=`${SED_CMD} -n '/#.*enabled_backends\s*/=' ${CINDER_CONF} | sed -n "1"p`
	line4=`${SED_CMD} -n '/\s*\[DEFAULT\]\s*/=' ${CINDER_CONF} | sed -n "1"p`

	if [ ! -z "${ENABLED_BACKENDS}" ];then
		ENABLED_BACKENDS=`echo "${ENABLED_BACKENDS}" | ${SED_CMD} 's/\s*,\s*/,/g'`
		ENABLED_BACKENDS=`echo "${ENABLED_BACKENDS}" | ${SED_CMD} 's/\t/\[tab\]/g'`
		ENABLED_BACKENDS=`echo "${ENABLED_BACKENDS}" | ${SED_CMD} 's/\s/\[space\]/g'`
		if [ ! -z "${line1}" ];then
			${SED_CMD} -i ''${line1}'s/enabled_backends\s*=.*/enabled_backends='${ENABLED_BACKENDS}'/' ${CINDER_CONF}
		elif [ ! -z "${line2}" ];then
			${SED_CMD} -i ''${line2}'a\enabled_backends='${ENABLED_BACKENDS}'\n' ${CINDER_CONF}
		elif [ ! -z "${line3}" ];then
			${SED_CMD} -i ''${line3}'a\enabled_backends='${ENABLED_BACKENDS}'\n' ${CINDER_CONF}
		elif [ ! -z "${line4}" ];then
			${SED_CMD} -i ''${line4}'a\\nenabled_backends='${ENABLED_BACKENDS}'\n' ${CINDER_CONF}
		else
			echo "ERROR: enabled_backends is set failed!"
			exit 1		
		fi
		line5=`${SED_CMD} -n '/'${ENABLED_BACKENDS}'/=' ${CINDER_CONF} | sed -n "1"p`
		${SED_CMD} -i ''${line5}'s/\[space\]/\ /g' ${CINDER_CONF}
		${SED_CMD} -i ''${line5}'s/\[tab\]/\t/g' ${CINDER_CONF}
		ENABLED_BACKENDS=""
	fi

	if [ ! -z "${VOLUME_BACKEND_NAME}" ];then
		echo volume_backend_name="${VOLUME_BACKEND_NAME}" >>${CINDER_CONF}
		VOLUME_BACKEND_NAME=""
	fi
	for line in ${params[@]}
	do
		echo ${line} >>${CINDER_CONF}
	done

	return 0
}

# create and output information to cinder_driver_config_file
OUTPUT_DRIVER_CONF()
{
	local section=$1
	local conf=$2
	local ip=$3
	local port=$4
	local user=$5
	local passwd=$6
	local poollist=$7
	local snappool=$8
	local iscsiip=$9

	${CAT_CMD} <<_EOT_ >"${conf}"
<?xml version='1.0' encoding='UTF-8'?>
<FUJITSU>
<EternusIP>${ip}</EternusIP>
<EternusPort>${port}</EternusPort>
<EternusUser>${user}</EternusUser>
<EternusPassword>${passwd}</EternusPassword>
_EOT_

	TMPIFS=${IFS}
	IFS=','
	for ip in ${iscsiip}
	do
		echo "<EternusISCSIIP>${ip}</EternusISCSIIP>" >> "${conf}"
	done

	for pool in ${poollist}
	do
		echo "<EternusPool>${pool}</EternusPool>" >> "${conf}"
	done

	echo "<EternusSnapPool>${snappool}</EternusSnapPool>" >> "${conf}"
	echo "</FUJITSU>" >> "${conf}"

	if [ $? -ne 0 ]; then
		# failed to create specified driver configuration file, return failure
		echo "Failed to create ${conf}" 
		exit 1
	fi

	IFS=${TMPIFS}

	if [ -f "${conf}" ]; then
		return 0
	else
		# failed to create specified driver configuration file, return failure
		echo "Failed to create ${conf}" 
		exit 1
	fi
}

# edit config files with manual
EDIT_CONFIG_MANUAL()
{
	local ans=1

	while [ ${ans} -eq 1 ]
	do
		local section=""
		local conf=""
		local ip=""
		local port=5988
		local user=""
		local passwd=""
		local iscsiip=""
		local pool=""
		local snappool=""
		local volume_backend_name=""

		TITLE "input driver configuration"

		while [ -z "${section}" ]
		do
			echo -n "section: "
			read -r section
		done

		while [ -z "${conf}" ]
		do
			echo -n "driver configuration path (.xml): "
			read -r conf
		done

		# check driver config file exsistence
		while [ -f "${conf}" ]
		do
			echo "SECTION: ${section}, specified driver configuration file (${conf}) has already existed"
			YES_NO "Overwrite the file (${conf})?"
			if [ $? -ne 0 ]; then
				conf=""
				while [ -z "${conf}" ]
				do
					echo -n "please re-enter driver configuration path (.xml): "
					read -r conf
				done
			else
				break
			fi
		done

		echo "input volume backend name, if you don't want to set please enter to skip"
		echo -n "volume_backend_name:"
		read -r volume_backend_name
		if [ ! -z "${volume_backend_name}" ]; then
			VOLUME_BACKEND_NAME="${volume_backend_name}"
		fi

		SELECT_PROTOCOL
		if [ $? -eq 1 ]; then #fc
			prtcl='fc'
		else
			prtcl='iscsi'
		fi

		while [ -z "${ip}" ]
		do
			echo -n "1)ETERNUS ip address for SMIS: "
			read -r ip
			ip=`echo ${ip} | ${TR_CMD} -d "' '|'\t'"`
		done

		# port
		echo -n "2)ETERNUS port number for SMIS: (default:5988) "
		read -r port
		port=`echo ${port} | ${TR_CMD} -d "' '|'\t'"`
		if [ -z "${port}" ]; then
			port=5988
		fi

		# usr
		while [ -z "${user}" ]
		do
			echo -n "3)ETERNUS username for SMIS: "
			read -r user
			user=`echo ${user} | ${TR_CMD} -d "' '|'\t'"`
		done

		# pass
		while [ -z "${passwd}" ]
		do
			echo -n "4)ETERNUS password for SMIS: "
			read -r passwd
			passwd=`echo ${passwd} | ${TR_CMD} -d "' '|'\t'"`
		done

		# pool
		while [ -z "${pool}" ]
		do
			echo -n "5)ETERNUS poolname (e.g. Pool01,Pool02,Pool03): "
			read -r pool
			pool=`echo "${pool}" | ${SED_CMD} 's/\s*,\s*/,/g'`
			pool=`echo "${pool}" | ${SED_CMD} 's/\t//g'`
		done

		# snappool
		echo -n "6)ETERNUS snappoolname: "
		read -r snappool
		if [ -z "${snappool}" ]; then
			snappool=`echo "${pool}" | ${CUT_CMD} -d ',' -f 1`
		fi
		snappool=`echo "${snappool}" | ${SED_CMD} 's/\t//g'`

		# ISCSIIP
		if [ ${prtcl} == 'iscsi' ]
		then
			while [ -z "${iscsiip}" ]
			do
				echo "7)iSCSI ip address (e.g. x.x.x.x,y.y.y.y,z.z.z.z): "
				read -r iscsiip
				iscsiip=`echo ${iscsiip} | ${TR_CMD} -d "' '|'\t'"`
			done
		fi

		echo ""
		echo "1)IP      : ${ip}"
		echo "2)PORT    : ${port}"
		echo "3)USR     : ${user}"
		echo "4)PASS    : ${passwd}"
		echo "5)POOL    : ${pool}"
		echo "6)SNAPPOOL    : ${snappool}"
		if [ ${prtcl} == 'iscsi' ]; then
			echo "7)iSCSIIP : ${iscsiip}"
		fi
		YES_NO "Are you sure?"
		ans=$?
	done

	# output informations to cinder.conf
	OUTPUT_CINDER_CONF "${section}" "${conf}"

	if [ $? -ne 0 ]; then
		return
	fi

	# output information to driver config file(xml)
	OUTPUT_DRIVER_CONF "${section}" "${conf}" "${ip}" "${port}" "${user}" "${passwd}" "${pool}" "${snappool}" "${iscsiip}"
	if [ $? -ne 0 ]; then
		return
	fi
}

# update config with manual
UPDATE_CONFIG_MANUAL()
{
	local loop_stop=0
	local ans=1
	local enabled_backends=""

	while [ ${ans} -eq 1 ]
	do
		echo "input enabled backends, if you don't want to change please enter to skip"
		echo -n "enabled backends:"
		read -r enabled_backends
		if [ ! -z "${enabled_backends}" ]; then
			echo "ENABLED_BACKENDS : ${enabled_backends}"
		fi
		YES_NO "Are you sure?"
		ans=$?
	done
	if [ ! -z "${enabled_backends}" ]; then
		ENABLED_BACKENDS=${enabled_backends}
	fi

	while [ ${loop_stop} -ne 1 ]
	do
		EDIT_CONFIG_MANUAL
		YES_NO "Continue to add configuration?"
		loop_stop=$?
	done
}

# edit config files silently
EDIT_CONFIG_SILENT()
{
	local section=$1
	local params=${*:2}
	local enabled="false"
	local prtcl=""
	local conf=""
	local enabled_backends=""
	local ip=""
	local port=5988
	local user=""
	local passwd=""
	local iscsiip=""
	local pool=""
	local snappool=""
	local others=()
	local i=0

	for param in ${params[@]}
	do
		key=`echo $param | ${CUT_CMD} -d '=' -f 1`
		value=`echo $param | ${CUT_CMD} -d '=' -f 2-`
		case ${key} in
			${STORAGE_ENABLED_KEY})  enabled=${value,,};; 
			${STORAGE_PRTCL_KEY})    prtcl=${value,,};;
			${STORAGE_CONF_KEY})     conf=${value};;
			${STORAGE_IP_KEY})       ip=${value};;
			${STORAGE_PORT_KEY})     port=${value};;
			${STORAGE_USR_KEY})      user=${value};;
			${STORAGE_PASSWD_KEY})   passwd=${value};;
			${STORAGE_ISCSIIP_KEY})  iscsiip=${value};;
			${STORAGE_POOL_KEY})     pool=${value};;
			${STORAGE_SNAPPOOL_KEY}) snappool=${value};;
			${STORAGE_BACKEND_KEY})  enabled_backends=${value};; 
			*)                       others+=(${param});;
		esac
	done
	
	if [ "${section}" == 'DEFAULT' ]; then
		if [ ! -z "${enabled_backends}" ]; then
			ENABLED_BACKENDS=${enabled_backends}
		fi
		return
	fi

	if [ "${enabled}" != 'true' ]; then
		echo "SECTION: ${section}, Skip"
		return
	fi
	
	if [ -z "${prtcl}" ] || [ -z "${conf}" ] || [ -z "${ip}" ] || [ -z "${port}" ] || \
		[ -z "${user}" ] || [ -z "${passwd}" ] || [ -z "${pool}" ]; then
		echo "SECTION: ${section}, Invalid Parameter. Please check each parameter."
		echo "Protocol : ${prtcl}"
		echo "Driver Configuration File : ${conf}"
		echo "IP Address: ${ip}"
		echo "PortNo: ${port}"
		echo "User: ${user}"
		echo "Password: ${passwd//?/'*'}"
		echo "PoolName: ${pool}"
		exit 1
	fi
	
	# check protocol type
	if [ "${prtcl}" != 'fc' ] && [ "${prtcl}" != 'iscsi' ]; then
		echo "SECTION: ${section}, Invalid Protocol: ${prtcl}"
		exit 1
	elif [ "${prtcl}" == 'iscsi' ] && [ -z "${iscsiip}" ]; then
		# when protocol is iscsi and iscsiip is null, return failure
		echo "SECTION: ${section}, iSCSI IP is not specified" 
		exit 1
	fi

	# when snappool is null, set poolname to it
	if [ -z "${snappool}" ]; then
		snappool=`echo ${pool} | ${CUT_CMD} -d ',' -f 1`
	fi

	# output informations to cinder.conf
	OUTPUT_CINDER_CONF "${section}" "${conf}" "${others[@]}"
	if [ $? -ne 0 ]; then
		return
	fi

	# check driver config file exsistence
	if [ -f "${conf}" ]; then
		echo "SECTION: ${section}, specified driver configuration file (${conf}) has already existed, Skip"
		return
	fi

	# output information to driver config file(xml)
	OUTPUT_DRIVER_CONF "${section}" "${conf}" "${ip}" "${port}" "${user}" "${passwd}" "${pool}" "${snappool}" "${iscsiip}"
	if [ $? -ne 0 ]; then
		return
	fi

	echo "SECTION: ${section}, Configure complete to ${CINDER_CONF} and ${conf}."
}

# update config silently
UPDATE_CONFIG_SILENT()
{
	local inst_cfg=$1
	local section=""
	local params=()
	local line=""

	OLDIFS=${IFS}
	IFS=$'\n'
	for rawline in `${CAT_CMD} ${inst_cfg}`
	do
		line=`echo ${rawline} | ${TR_CMD} -d "' '|'\t'"`

		# if line is null or comment, then skip
		echo ${line} | ${GREP_CMD} "#\|^$">/dev/null
		if [ $? -eq 0 ]; then
			continue
		fi

		## if line is section line, then update section information
		echo ${line} | ${GREP_CMD} "^\[" | ${GREP_CMD} "\]$" >/dev/null
		if [ $? -eq 0 ]; then
			# flush previous section information
			if [ -n "${section}" ]; then
				EDIT_CONFIG_SILENT "${section}" "${params[@]}"
			fi

			# initialize section information
			section=`echo ${line} | sed 's/^\[\|\]$//g'`
			params=()
		else
			params+=($line)
		fi

	done

	# flush previous section information
	if [ -n "${section}" ]; then
		EDIT_CONFIG_SILENT "${section}" "${params[@]}"
	fi
	IFS=${OLDIFS}
}

#---------------
# DELETE SubFunctions
#---------------

# delete files
DELETE_FILE()
{
	inst_dir=$1
	TITLE "delete driver files"
	echo "remove fujitsu from ${inst_dir}"
	${RM_CMD} -r ${inst_dir}fujitsu 2>/dev/null
}

# INSTALL
INSTALL()
{
	local inst_cfg=$1
	local inst_dir=""

	TITLE "Install Start"
	
	if [ -z "$1" ]; then
		echo "Install start with interactive input"
	else
		echo "Install start with ${inst_cfg}"
	fi

	inst_dir=`FIND_INSTDIR 2>/dev/null`
	if [ $? -ne 0 ]; then
		PR "Cannot find directory for cinder volume drivers" "ERROR"
		exit 1
	fi

	CHECK_INSTALLED "${inst_dir}" "${inst_cfg}"
	PUT_FILE "${inst_dir}"
	UPDATE_CONFIG "${inst_cfg}"
	
# rootwrap
	TITLE "set rootwrap for eternus_dx_cli"
	if [ ! -d /etc/cinder/rootwrap.d ]
	then
		PR "create directory[/etc/cinder/rootwrap.d]"
		${MKDIR_CMD} /etc/cinder/rootwrap.d
	fi
	if [ -f /etc/cinder/rootwrap.d/volume.filters ]
	then
		count=`${GREP_CMD} -c fujitsu/extension/eternus_dx_cli /etc/cinder/rootwrap.d/volume.filters`
		if [ ${count} -eq 0 ]
		then
			PR "add eternus_dx_cli [/etc/cinder/rootwrap.d/volume.filters]"
			echo "eternus_dx_cli: CommandFilter, ${inst_dir}fujitsu/extension/eternus_dx_cli, root" >> /etc/cinder/rootwrap.d/volume.filters
		fi
	else
		PR "create file [/etc/cinder/rootwrap.d/volume.filters]"
		echo "[Filters]" > /etc/cinder/rootwrap.d/volume.filters
		echo "eternus_dx_cli: CommandFilter, ${inst_dir}fujitsu/extension/eternus_dx_cli, root" >> /etc/cinder/rootwrap.d/volume.filters
	fi

# create ssh key
	TITLE "create ssh key for eternus"
	if [ ! -f /root/.ssh/eternus ]
	then
		PR "create ssh key"
		${SSHKEYGEN_CMD} -t rsa -N "" -f /root/.ssh/eternus
	fi
	PR "create eternus.ietf"
	${SSHKEYGEN_CMD} -e -f /root/.ssh/eternus.pub > ./eternus.ietf
}

# UPDATE
UPDATE_CONFIG()
{
	local inst_cfg=$1

	TITLE "Update Cinder Configuration"
	if [ -z "${inst_cfg}" ]; then
		UPDATE_CONFIG_MANUAL
	else
		UPDATE_CONFIG_SILENT "${inst_cfg}"
	fi
}

# UNINSTALL
UNINSTALL()
{
	local inst_cfg=$1

	TITLE "Uninstall Start"

	inst_dir=`FIND_INSTDIR 2>/dev/null`
	if [ $? -ne 0 ]; then
		PR "Cannot find directory for cinder volume drivers" "ERROR"
		exit 1
	fi

	DELETE_FILE ${inst_dir}
}
#---------------
# Main
#---------------
inst_cmd=$1
inst_cfg=$2
PR_PRODUCT_INFO

# find multi exec
find_multi_exec=`pgrep -fo $0`
if [ $$ != ${find_multi_exec} ]; then
	PR "Another user is running this installer."
	PR "The installation is ended."
	exit 1
fi

# check login user
inst_user=`${LS_CMD} -ld ${inst_dir}|${AWK_CMD} '{print $3}'`
if [ ${inst_user} != `${WHOAMI_CMD}` ]; then
	PR "Please execute installer after re-log in as a user \"${inst_user}\" where you installed cinder." "ERROR"
	exit 1
fi

case "${inst_cmd}" in
	"install")
		INSTALL "${inst_cfg}"
		;;
	"update")
		UPDATE_CONFIG "${inst_cfg}"
		;;
	"uninstall")
		UNINSTALL
		;;
	*)
		PR_ERROR_USAGE
		;;
esac

echo ""
PR "${inst_cmd^} is FINISHED"
PR "Please reboot cinder services to reflect configuration done by this installer."

exit 0
