#!/bin/bash
# $Date:: 2015-08-31 08:07:56 +0800#$
# $Revision: 155 $
# Usage: sh setup_fujitsu_image.sh <image_id> <volume_size> [ <volume_type> ]

# include
if [ -z "${OS_USERNAME}" ] || [ -z "${OS_TENANT_NAME}" ] ||
   [ -z "${OS_PASSWORD}" ] || [ -z "${OS_AUTH_URL}" ]; then
   
  source ./keystonerc_fujitsu_eternus
fi

# function
function print_error(){
  local msg=$1
  local efile=$2
  local imgid=$3
  local vsize=$4
  local vtype=$5

  echo "Setup Image Volume: fail"
  
  echo -n "Message:  "
  echo "${msg}"

  echo -n "Error Message:  "
  cat ${efile}

  echo "Check following parameters"
  echo "Image ID: ${imgid}"
  echo "Volume Size: ${vsize}"

  if [ -n "${vtype}" ]; then
    echo "Volume Type: ${vtype}"
  fi
}

# initialize
image_id=$1
volume_size=$2
wait_time=10
volume_type_opt=""
readonly error_message_file="./error_message"

# main
if [ -z "${image_id}" ] || [ -z "${volume_size}" ]; then
    echo "usage: sh setup_fujitsu_image.sh <image_id> <volume_size> [ <volume_type> ]"
    exit 1
fi

digit_check=`echo -n ${volume_size} | sed 's/[0-9]//g' `
if [ -n "$digit_check" ]; then
    echo "Error: Non-digital value was specified to volume_size"
    echo "usage: sh setup_fujitsu_image.sh <image_id> <volume_size> [ <volume_type> ]"
    exit 1
fi

if [ -n "$3" ]; then
    volume_type_opt="--volume-type $3"
fi

trap 'rm -f ${error_message_file}' INT TERM EXIT

echo "Setup Image Volume: start"
touch ${error_message_file}

result_message=`cinder create --image-id ${image_id}  --display-name ETERNUS_TMP_Volume_${image_id} \
                ${volume_type_opt} --metadata fujitsu_image_setup=True -- ${volume_size} 2>${error_message_file}`
result_code=$?

if [ ${result_code} -ne 0 ]; then
  print_error "${result_message}" ${error_message_file} ${image_id} ${volume_size} $3
  exit 1
fi

create_volume_id=`echo "$result_message" | grep ' id ' | awk '{print $4}'`

while true
do
    sleep ${wait_time}
    status=`cinder list | grep ${create_volume_id} |  awk '{print $4}'`
    if [ -z "${status}" ]; then
        break
    elif [ "${status}" == "creating" ]; then
        continue
    elif [ "${status}" == "downloading" ]; then
        wait_time=60
        continue
    elif [ "${status}" == "error" ]; then
        cinder force-delete ${create_volume_id}
        echo "Setup Image Volume: fail"
        echo "Check configuration of cinder.conf and free space on target StoragePool"
        echo "For error details, please read cinder logs"
        break
    else
        cinder force-delete ${create_volume_id}
        echo "Setup Image Volume: finish"
        break
    fi
done
